#!/usr/bin/env python3

# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.


"""Tectonic: An academic Cyber Range."""

import os
import pprint
import re
import logging
from configparser import ConfigParser
import traceback
from pathlib import Path, PurePosixPath
from collections import OrderedDict

import click

from tectonic.deployment_aws import AWSDeployment
from tectonic.deployment_libvirt import LibvirtDeployment
from tectonic.deployment_docker import DockerDeployment
from tectonic.description import Description
from tectonic.ansible import Ansible
from tectonic.instance_type import InstanceType
from tectonic.instance_type_aws import InstanceTypeAWS
from tectonic.utils import create_table



def log(msg):
    click.echo(msg)


def debug(msg):
    if DEBUG:
        click.echo(msg)


# The instances can be specified as a list, a range, or a combination of both:
#    --instances=1,3,8
#    --instances=1-5
#    --insntaces=13,16-24,25
# A comma on its own can be used to specify an empty list of instances.
class NumberRangeParamType(click.ParamType):
    """ A parameter that works similar to :data:`click.INT`, but accepts
    ranges of integers as well.  For instance, ``1-3,5,7-10`` would result
    in a list of four integers.
    """
    name = "NUMBER_RANGE"

    def convert(self, value, param, ctx):
        if isinstance(value, list):
            return sorted(set(value))

        try:
            values = []
            for i in value.split(","):
                m = re.search(r'^\d+$', i)
                if m:
                    values += [int(m.group(0))]
                else:
                    m = re.search(r'^(\d+)(\.\.|-)(\d+)$', i)
                    if m:
                        values += list(range(int(m.group(1)), int(m.group(3)) + 1))
                    elif i == "":
                        pass
                    else:
                        self.fail(f"Cannot parse {i}.", param, ctx)
            return sorted(set(values))
        except Exception as e:
            self.fail(f"{value!r} is not a valid range. {str(e)}", param, ctx)


class TectonicException(Exception):
    pass


def load_config(ctx, param, config_file):
    if config_file is None:
        return {}
    cfg = ConfigParser()
    cfg.read(config_file)
    options = {}

    # flatten the config file
    for section in ["config", "aws", "libvirt", "docker", "elastic", "caldera"]:
        try:
            options = options | dict(cfg[section])
        except KeyError:
            pass

    ctx.default_map = options


NUMBER_RANGE = NumberRangeParamType()
CONTEXT_SETTINGS = {"help_option_names": ['-h', '--help']}


def range_to_str(r):
    """Returns an appropriate text describing the range."""
    if not r:
        return ""

    s = []
    it = iter(r)
    prev = next(it)
    accum = [str(prev)]
    for n in it:
        if n == prev + 1:
            accum.append(str(n))
        else:
            if len(accum) > 2:
                s.append(f"from {accum[0]} to {accum[-1]}")
            else:
                s += accum
            accum = [str(n)]
        prev = n
    if len(accum) > 2:
        s.append(f"from {accum[0]} to {accum[-1]}")
    else:
        s += accum

    if len(s) > 1:
        return ", ".join(s[:-1]) + ", and " + s[-1]

    return s[0]


def confirm_machines(ctx, instances, guest_names, copies, action):
    """Prompt the user for confirmation to perform ACTION to machines."""
    print_instances = True

    if instances:
        instances = list(
            filter(lambda i: i <= ctx.obj["description"].instance_number, instances)
        )

    if not instances or len(instances) == ctx.obj["description"].instance_number:
        instances_msg = "all instances"
    elif len(instances) == 1:
        instances_msg = f"instance {instances[0]}"
    else:
        instances_msg = f"instances {range_to_str(instances)}"

    if not guest_names:
        machines_msg = "all machines"
    else:
        machines = []

        # Remove duplicates
        guest_names = list(OrderedDict.fromkeys(guest_names))
        if "teacher_access" in guest_names:
            machines += ["the teacher access"]
            guest_names.remove("teacher_access")
        if "student_access" in guest_names:
            machines += ["the student access"]
            guest_names.remove("student_access")
        if "packetbeat" in guest_names:
            machines += ["the packetbeat"]
            guest_names.remove("packetbeat")

        if not guest_names:
            print_instances = False

        for guest_name in guest_names:
            if ctx.obj["description"].get_guest_copies(guest_name) == 1:
                if not copies or 1 in copies:
                    machines += [f"the {guest_name}"]
            else:
                if not copies:
                    machines += [f"all copies of the {guest_name}"]
                else:
                    guest_copies = list(
                        filter(
                            lambda c: c
                            <= ctx.obj["description"].get_guest_copies(guest_name),
                            copies,
                        )
                    )
                    if len(guest_copies) == 1:
                        machines += [f"copy {guest_copies[0]} of the {guest_name}"]
                    else:
                        machines += [
                            f"copies {range_to_str(guest_copies)} of the {guest_name}"
                        ]
        if len(machines) > 1:
            machines_msg = ", ".join(machines[:-1]) + " and " + machines[-1]
        elif len(machines) == 1:
            machines_msg = machines[0]
        else:
            return

    if print_instances:
        click.echo(f"{action} {machines_msg}, on {instances_msg}.")
    else:
        click.echo(f"{action} {machines_msg}.")
    click.confirm("Continue?", abort=True)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    # TODO: Try to find the file in the correct locations.
    callback=load_config,
    is_eager=True,
    expose_value=False,
    help="Read option defaults from specified INI file.",
    show_default=True,
)
@click.option(
    "--debug/--no-debug", default=False, help="Show debug messages during execution."
)
@click.option(
    "--platform",
    type=click.Choice(["aws", "libvirt", "docker"], case_sensitive=False),
    default="aws",
    help="Deploy the cyber range to this platform.",
)
@click.option(
    "--lab_repo_uri",
    "-u",
    required=True,
    help="URI to a lab repository. Labs to deploy will be searched in this repo.",
)
@click.option(
    "--aws_region",
    "-r",
    default="us-east-1",
    show_default=True,
    help="AWS region to use for deployment.",
)
@click.option(
    "--aws_default_instance_type",
    "-t",
    default="t2.micro",
    show_default=True,
    help="Default EC2 instance type. Can be overwritten in CR guest configuration.",
)
@click.option(
    "--ssh_public_key_file",
    "-p",
    type=click.Path(dir_okay=False),
    default="~/.ssh/id_rsa.pub",
    show_default=True,
    help="SSH pubkey to be used to connect to machines for configuration.",
)
@click.option(
    "--ansible_ssh_common_args",
    "-s",
    default="-o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -o ControlMaster=auto -o ControlPersist=3600 ",
    help="SSH extra connection options for connection to the CR machines.",
)  # TODO: check default value
@click.option(
    "--packetbeat_vlan_id",
    "-v",
    default="1",
    help="VLAN ID used in traffic mirrorring.",
)
@click.option(
    "--packetbeat_policy_name",
    "-n",
    default="Packetbeat",
    help="Packetbeat policy name.",
)
@click.option(
    "--network_cidr_block",
    "-b",
    default="10.0.0.0/16",
    help="CIDR block for the whole lab network.",
)
@click.option(
    "--configure_dns",
    default=True,
    show_default=True,
    help="Configure internal DNS for instances.",
)
@click.option(
    "--teacher_access",
    type=click.Choice(["host", "endpoint"], case_sensitive=False),
    default="host",
    help="Type of teacher access to configure.",
)
@click.option(
    "--elastic_stack_version",
    default="8.14.3",
    help="Elastc version",
)
@click.option(
    "--gitlab_backend_url",
    required=False,
    help="Gitlab terraform state url",
)
@click.option(
    "--gitlab_backend_username",
    envvar="GITLAB_USERNAME",
    required=False,
    help="Gitlab Username",
)
@click.option(
    "--gitlab_backend_access_token",
    envvar="GITLAB_ACCESS_TOKEN",
    required=False,
    help="Gitlab Access Token",
)
@click.option(
    "--packer_executable_path",
    required=False,
    help="Packer executable path",
    default="packer",
)
@click.option(
    "--libvirt_uri",
    default="qemu:///system",
    show_default=True,
    help="URI to connect to server, if using libvirt",
)
@click.option(
    "--libvirt_storage_pool",
    default="default",
    help="Name of the libvirt storage pool where images will be created.",
)
@click.option(
    "--libvirt_student_access",
    type=click.Choice(["port_forwarding", "bridge"], case_sensitive=False),
    default="port_forwarding",
    help="How student access should be configured: port forwarding in the libvirt server (port_forwarding) or adding "
         "NICs to the entry points in a bridged network (bridge).",
)
@click.option(
    "--libvirt_bridge",
    help="Name of the libvirt server bridge to use for student access.",
)
@click.option(
    "--libvirt_external_network",
    default="192.168.0.0/25",
    help="CIDR block of the external bridged network, if appropriate. Static IP addresses are assigned sequentially "
         "to lab entry points.",
)
@click.option(
    "--libvirt_bridge_base_ip",
    default=10,
    help="Starting IP from which to sequentially assign the entry points IPs. With default values, the first entry "
         "point would get 192.168.44.11, and so on.",
)
@click.option(
    "--proxy",
    required=False,
    help="Guest machines proxy configuration URI for libvirt.",
)
@click.option(
    "--endpoint_policy_name",
    default="Endpoint",
    help="Agent policy name.",
)
@click.option(
    "--user_install_packetbeat",
    default="tectonic",
    help="User used to install Packetbeat",
)
@click.option(
    "--internet_network_cidr_block",
    default="192.168.4.0/24",
    help="CIDR internet network",
)
@click.option(
    "--services_network_cidr_block",
    default="192.168.5.0/24",
    help="CIDR services network",
)
@click.option(
    "--keep_ansible_logs/--no-keep_ansible_logs",
    default="False",
    help="Keep Ansible logs on managed hosts.",
)
@click.option(
    "--docker_uri",
    default="unix:///var/run/docker.sock",
    show_default=True,
    help="URI to connect to server, if using docker",
)
@click.option(
    "--caldera_version",
    default="latest",
    help="Caldera version.",
)
@click.option(
    "--docker_dns",
    default="8.8.8.8",
    required=False,
    help="DNS to use for internet networks on Docker",
)
@click.option(
    "--ansible_forks",
    default="10",
    required=False,
    help="Number of parallel connection for Ansible",
)
@click.option(
    "--ansible_pipelining/--no-ansible_pipelining",
    default="False",
    help="Enable pipelining for Ansible",
)
@click.option(
    "--ansible_timeout",
    default="10",
    required=False,
    help="Timeout for Ansible connection.",
)
@click.argument("lab_edition_file", type=click.Path(exists=True, dir_okay=False))
@click.pass_context
def tectonic(
    ctx,
    platform,
    lab_repo_uri,
    aws_region,
    aws_default_instance_type,
    ssh_public_key_file,
    ansible_ssh_common_args,
    debug,
    packetbeat_vlan_id,
    packetbeat_policy_name,
    network_cidr_block,
    configure_dns,
    teacher_access,
    elastic_stack_version,
    gitlab_backend_url,
    gitlab_backend_username,
    gitlab_backend_access_token,
    packer_executable_path,
    libvirt_uri,
    libvirt_storage_pool,
    libvirt_student_access,
    libvirt_bridge,
    libvirt_external_network,
    libvirt_bridge_base_ip,
    proxy,
    lab_edition_file,
    endpoint_policy_name,
    user_install_packetbeat,
    internet_network_cidr_block,
    services_network_cidr_block,
    keep_ansible_logs,
    docker_uri,
    caldera_version,
    docker_dns,
    ansible_forks,
    ansible_pipelining,
    ansible_timeout
):
    """Deploy or manage a cyber range according to LAB_EDITION_FILE."""
    logfile = PurePosixPath(lab_edition_file).parent.joinpath("tectonic.log")
    loglevel = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(filename=logfile, encoding="utf-8", level=loglevel)

    ctx.ensure_object(dict)

    instance_type = None
    if platform == "aws":
        instance_type = InstanceTypeAWS(aws_default_instance_type)
    else:
        instance_type = InstanceType()

    ctx.obj["description"] = Description(
        lab_edition_file,
        platform,
        lab_repo_uri,
        teacher_access,
        configure_dns,
        ssh_public_key_file,
        ansible_ssh_common_args,
        aws_region,
        aws_default_instance_type,
        network_cidr_block,
        packetbeat_policy_name,
        packetbeat_vlan_id,
        elastic_stack_version,
        libvirt_uri,
        libvirt_storage_pool,
        libvirt_student_access,
        libvirt_bridge,
        libvirt_external_network,
        libvirt_bridge_base_ip,
        proxy,
        instance_type,
        endpoint_policy_name,
        user_install_packetbeat,
        internet_network_cidr_block,
        services_network_cidr_block,
        keep_ansible_logs,
        docker_uri,
        caldera_version,
        docker_dns,
        ansible_forks,
        ansible_pipelining,
        ansible_timeout
    )

    if platform == "aws":
        ctx.obj["deployment"] = AWSDeployment(
            ctx.obj["description"],
            gitlab_backend_url=gitlab_backend_url,
            gitlab_backend_username=gitlab_backend_username,
            gitlab_backend_access_token=gitlab_backend_access_token,
            packer_executable_path=packer_executable_path,
        )
    elif platform == "libvirt":
        ctx.obj["deployment"] = LibvirtDeployment(
            ctx.obj["description"],
            gitlab_backend_url=gitlab_backend_url,
            gitlab_backend_username=gitlab_backend_username,
            gitlab_backend_access_token=gitlab_backend_access_token,
            packer_executable_path=packer_executable_path,
        )
    elif platform == "docker":
        ctx.obj["deployment"] = DockerDeployment(
            ctx.obj["description"],
            gitlab_backend_url=gitlab_backend_url,
            gitlab_backend_username=gitlab_backend_username,
            gitlab_backend_access_token=gitlab_backend_access_token,
            packer_executable_path=packer_executable_path,
        )
    else:
        raise Exception(f"Unsupported platform {platform}.")

    ctx.obj["ansible"] = Ansible(ctx.obj["deployment"])

    if elastic_stack_version == "latest":
        ctx.obj["description"].set_elastic_stack_version(ctx.obj["deployment"].get_elastic_latest_version())

    if caldera_version == "latest":
        ctx.obj["description"].set_caldera_version("master")


@tectonic.command()
@click.pass_context
@click.option(
    "--images/--no-images",
    default=False,
    show_default=True,
    help="Whether to create base images for each guest in the lab.",
)
@click.option(
    "--instances", "-i", help="Range of instances to list.", type=NUMBER_RANGE
)
@click.option(
    "--packetbeat_image/--no-packetbeat_image",
    default=False,
    show_default=True,
    help="Whether to create the base image for packetbeat.",
)
@click.option(
    "--elastic_image/--no-elastic_image",
    default=False,
    show_default=True,
    help="Whether to create the base image for ELK.",
)
@click.option(
    "--caldera_image/--no-caldera_image",
    default=False,
    show_default=True,
    help="Whether to create the base image for Caldera.",
)
@click.option(
    "--force",
    "-f",
    help="Force the deployment of instances without a confirmation prompt.",
    is_flag=True,
)
def deploy(ctx, images, instances, packetbeat_image, elastic_image, caldera_image, force):
    """Deploy the cyber range."""
    if not force:
        confirm_machines(ctx, instances, guest_names=None, copies=None, action="Deploying")

    if images:
        _create_images(ctx, packetbeat_image, elastic_image, caldera_image, True)

    ctx.obj["deployment"].deploy_infraestructure(instances)

    _info(ctx)


@tectonic.command()
@click.pass_context
@click.option(
    "--images/--no-images",
    default=False,
    show_default=True,
    help="Whether to destroy the base images of each guest in the lab.",
)
@click.option(
    "--packetbeat_image/--no-packetbeat_image",
    default=False,
    show_default=True,
    help="Whether to destroy the base image for packetbeat.",
)
@click.option(
    "--elastic_image/--no-elastic_image",
    default=False,
    show_default=True,
    help="Whether to destroy the base image for ELK.",
)
@click.option(
    "--caldera_image/--no-caldera_image",
    default=False,
    show_default=True,
    help="Whether to destroy the base image for Caldera.",
)
@click.option(
    "--instances", "-i", help="Range of instances to list.", type=NUMBER_RANGE
)
@click.option(
    "--force",
    "-f",
    help="Force the destruction of instances without a confirmation prompt.",
    is_flag=True,
)
def destroy(ctx, images, instances, packetbeat_image, elastic_image, caldera_image, force):
    """Delete and destroy all resources of the cyber range."""

    if not force:
        confirm_machines(
            ctx, instances, guest_names=None, copies=None, action="Destroying"
        )

    ctx.obj["deployment"].destroy_infraestructure(instances)

    if instances is None:
        if images:
            click.echo("Destroying base images...")
            ctx.obj["deployment"].delete_cr_images()

        if (packetbeat_image and ctx.obj["description"].platform == "aws")  or elastic_image or caldera_image:
            click.echo("Destroying services base image...")
            services = {
                "packetbeat": packetbeat_image and ctx.obj["description"].platform == "aws",
                "elastic": elastic_image,
                "caldera": caldera_image
            }
            ctx.obj["deployment"].delete_services_images(services)

@tectonic.command()
@click.pass_context
@click.option(
    "--packetbeat/--no-packetbeat",
    default=False,
    show_default=True,
    help="Whether to create the base image for packetbeat.",
)
@click.option(
    "--elastic/--no-elastic",
    default=False,
    show_default=True,
    help="Whether to create the base image for Elastic on-prem.",
)
@click.option(
    "--machines/--no-machines",
    default=True,
    show_default=True,
    help="Whether to create the scenario base images.",
)
@click.option(
    "--caldera/--no-caldera",
    default=False,
    show_default=True,
    help="Whether to create the base image for Caldera on-prem.",
)
@click.option(
    "--guests",
    "-g",
    multiple=True,
    type=click.STRING,
    help="Name of guests to list.",
)
def create_images(ctx, packetbeat, elastic, caldera, machines, guests):
    """Create lab base images."""
    ctx.obj["description"].parse_machines(
        instances=None, guests=guests, copies=None, only_instances=True
    )
    _create_images(ctx, packetbeat, elastic, caldera, machines, guests)


def _create_images(ctx, packetbeat, elastic, caldera, machines, guests=None):
    if (packetbeat and ctx.obj["description"].platform == "aws") or elastic or caldera:
        click.echo("Creating services images ...")
        services = {
            "packetbeat": packetbeat and ctx.obj["description"].platform == "aws",
            "elastic": elastic,
            "caldera": caldera
        }
        ctx.obj["deployment"].create_services_images(services)

    if machines:
        click.echo("Creating base images...")
        ctx.obj["deployment"].create_cr_images(guests)


@tectonic.command(name="list")
@click.pass_context
@click.option(
    "--instances", "-i", help="Range of instances to list.", type=NUMBER_RANGE
)
@click.option(
    "--guests", "-g", help="Name of guests to list.", multiple=True, type=click.STRING
)
@click.option("--copies", "-c", help="Number of copy to list.", type=NUMBER_RANGE)
def list_instances(ctx, instances, guests, copies):
    """Print information and state of the cyber range resources."""
    click.echo("Getting Cyber Range status...")
    result = ctx.obj["deployment"].list_instances(instances, guests, copies)
    click.echo(result)
    click.echo("Getting Services status...")
    result = ctx.obj["deployment"].get_services_status()
    click.echo(result)

@tectonic.command()
@click.pass_context
@click.option(
    "--instances", "-i", help="Range of instances to start.", type=NUMBER_RANGE
)
@click.option(
    "--guests", "-g", help="Name of guests to start.", multiple=True, type=click.STRING
)
@click.option("--copies", "-c", help="Number of copy to start.", type=NUMBER_RANGE)
@click.option(
    "--force",
    "-f",
    help="Force the start of instances without a confirmation prompt.",
    is_flag=True,
)
@click.option(
    "--services",
    "-s",
    help="Start services components.",
    is_flag=True,
)
def start(ctx, instances, guests, copies, force, services):
    """Start (boot up) machines in the cyber range."""
    if not force:
        confirm_machines(ctx, instances, guests, copies, "Start")
    ctx.obj["deployment"].start(instances, guests, copies, services)


@tectonic.command()
@click.pass_context
@click.option(
    "--instances", "-i", help="Range of instances to shutdown.", type=NUMBER_RANGE
)
@click.option(
    "--guests",
    "-g",
    help="Name of guests to shutdown.",
    multiple=True,
    type=click.STRING,
)
@click.option(
    "--copies", "-c", help="Number of copy to shutdown.", multiple=True, type=click.INT
)
@click.option(
    "--force",
    "-f",
    help="Force the shut down of instances without a confirmation prompt.",
    is_flag=True,
)
@click.option(
    "--services",
    "-s",
    help="Shut down services components.",
    is_flag=True,
)
def shutdown(ctx, instances, guests, copies, force, services):
    """Shutdown machines in the cyber range."""
    if not force:
        confirm_machines(ctx, instances, guests, copies, "Shut down")
    ctx.obj["deployment"].shutdown(instances, guests, copies, services)

@tectonic.command()
@click.pass_context
@click.option(
    "--instances", "-i", help="Range of instances to reboot.", type=NUMBER_RANGE
)
@click.option(
    "--guests", "-g", help="Name of guests to reboot.", multiple=True, type=click.STRING
)
@click.option("--copies", "-c", help="Number of copy to reboot.", type=NUMBER_RANGE)
@click.option(
    "--force",
    "-f",
    help="Force the reboot of instances without a confirmation prompt.",
    is_flag=True,
)
@click.option(
    "--services",
    "-s",
    help="Reboot services components.",
    is_flag=True,
)
def reboot(ctx, instances, guests, copies, force, services):
    """Reboot machines in the cyber range."""
    if not force:
        confirm_machines(ctx, instances, guests, copies, "Reboot")
    ctx.obj["deployment"].reboot(instances, guests, copies, services)


@tectonic.command()
@click.pass_context
@click.option(
    "--instance", "-i", help="Number of instance to connect.", type=NUMBER_RANGE
)
@click.option(
    "--guest", "-g", help="Name of guest to connect.", multiple=True, type=click.STRING
)
@click.option("--copy", "-c", help="Number of copy to connect.", type=NUMBER_RANGE)
@click.option(
    "--username",
    "-u",
    help="Username for connecting to the machine.",
    type=click.STRING,
)
def console(ctx, instance, guest, copy, username):
    """Connect to a machine in the cyber range and get a console."""
    machines_to_connect = ctx.obj["description"].parse_machines(
        instance or [], guest, copy, False
    )
    if len(machines_to_connect) > 1:
        raise TectonicException("You must specify only one machine to connect.")

    machine_to_connect = machines_to_connect[0]
    click.echo(f"Connecting to machine {machine_to_connect}...")
    ctx.obj["deployment"].connect_to_instance(machine_to_connect, username)


@tectonic.command()
@click.pass_context
@click.option(
    "--instances",
    "-i",
    help="Run ansible only on this range of instances.",
    type=NUMBER_RANGE,
)
@click.option(
    "--guests",
    "-g",
    help="Run ansible only on these guests with this base name.",
    multiple=True,
    type=click.STRING,
)
@click.option(
    "--copies",
    "-c",
    help="Run ansible only on guests with this copy number.",
    type=NUMBER_RANGE,
)
@click.option(
    "--username",
    "-u",
    help="Username for connecting to the machine.",
    type=click.STRING,
)
@click.option(
    "--playbook",
    "-p",
    help="Playbook to run. If not provided, the lab default after-clone playbook will be used.",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--force",
    "-f",
    help="Run the playbook on selected instances without a confirmation prompt.",
    is_flag=True,
)
def run_ansible(ctx, instances, guests, copies, username, playbook, force):
    """Run an ansible playbook in the selected INSTANCE."""
    ctx.obj["description"].parse_machines(
        instances, guests, copies, only_instances=False
    )
    if not force:
        confirm_machines(ctx, instances, guests, copies, "Applying ansible playbook to")
    ctx.obj["ansible"].run(
        instances=instances, guests=guests, copies=copies, only_instances=False, username=username, playbook=playbook
    )


@tectonic.command()
@click.pass_context
@click.option(
    "--instances", "-i", help="Range of instances to configure student access credentials.", type=NUMBER_RANGE
)
@click.option(
    "--force",
    "-f",
    help="Apply the configuration to the selected machines without a confirmation prompt.",
    is_flag=True,
)
def student_access(ctx, instances, force):
    """Create users for the students.
    Users are created on all entry points (and the student access
    host, if appropriate). Credentials can be public SSH keys and/or
    autogenerated passwords.
    """

    entry_points = [ base_name for base_name, guest in ctx.obj["description"].guest_settings.items()
                     if guest and guest.get("entry_point") ]

    if ctx.obj["description"].platform == "aws":
        entry_points += ["student_access"]

    if not force:
        confirm_machines(ctx, instances, entry_points, None, "Creating credentials on")

    click.echo("Configuring student access...")
    ctx.obj["deployment"].student_access(instances)

    _print_student_passwords(ctx)

@tectonic.command()
@click.pass_context
def info(ctx):
    """Show cyber range information."""
    _info(ctx)

def _info(ctx):
    click.echo("Getting Cyber Range information...")
    result = ctx.obj["deployment"].get_cyberrange_data()
    click.echo(result)
    _print_student_passwords(ctx)
    
def _print_student_passwords(ctx):
    """Print the generated student passwords, if create_student_passwords is True.
    """
    if ctx.obj["description"].create_student_passwords:
        users = ctx.obj["deployment"].get_student_access_users()
        click.echo("\nStudent users:")
        rows = []
        headers = ["Username", "Password"]
        for username, user in users.items():
            rows.append([username, user['password']])
        table = create_table(headers, rows)
        click.echo(table)


@tectonic.command()
@click.pass_context
@click.option(
    "--instances", "-i", help="Range of instances to recreate.", type=NUMBER_RANGE
)
@click.option(
    "--guests",
    "-g",
    help="Name of guests to recreate.",
    multiple=True,
    type=click.STRING,
)
@click.option(
    "--copies",
    "-c",
    help="Number of copy to recreate.",
    type=NUMBER_RANGE,
)
@click.option(
    "--force",
    "-f",
    help="Recreate the selected machines without a confirmation prompt.",
    is_flag=True,
)
def recreate(ctx, instances, guests, copies, force):
    """Recreate instances."""
    if not force:
        confirm_machines(ctx, instances, guests, copies, "Recreate")
    ctx.obj["deployment"].recreate(instances, guests, copies)

if __name__ == "__main__":
    obj = {}
    try:
        tectonic(obj)
    except Exception as e:
        if obj["debug"]:
            traceback.print_exc()
        else:
            click.echo(f"Error: {e}")

@tectonic.command()
@click.pass_context
@click.option(
    "--instances", "-i", help="Range of instances to list.", type=NUMBER_RANGE
)
@click.option(
    "--directory",
    "-d",
    help="path of the directory where to create the parameters file.",
    type=click.STRING,
)
def show_parameters(ctx, instances, directory):
    """Generate parameters for instances"""
    click.echo("Getting parameters")
    ctx.obj["deployment"].get_parameters(instances, directory)




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import ipaddress
import math
import tempfile
import random
import yaml
from zipfile import ZipFile
from pathlib import Path
import re
import os
import json

from tectonic.constants import OS_DATA
from tectonic.instance_type import InstanceType
from tectonic.utils import read_files_in_directory

class DescriptionException(Exception):
    pass


class Description(object):
    """Class to represent a lab description."""

    institution = ""
    lab_name = ""
    instance_number = 0
    default_os = "ubuntu22"
    deploy_elastic = False
    elastic_deployment = {}
    guest_settings = {}
    topology = []

    def __init__(
        self,
        path,
        platform,
        lab_repo_uri,
        teacher_access,
        configure_dns,
        ssh_public_key_file,
        ansible_ssh_common_args,
        aws_region,
        aws_default_instance_type,
        network_cidr_block,
        packetbeat_policy_name,
        packetbeat_vlan_id,
        elastic_stack_version,
        libvirt_uri,
        libvirt_storage_pool,
        libvirt_student_access,
        libvirt_bridge,
        libvirt_external_network,
        libvirt_bridge_base_ip,
        proxy,
        instance_type,
        endpoint_policy_name,
        user_install_packetbeat,
        internet_network_cidr_block,
        services_network_cidr_block,
        keep_ansible_logs,
        docker_uri,
        caldera_version,
        docker_dns,
        ansible_forks,
        ansible_pipelining,
        ansible_timeout
    ):
        """Create a Description object.

        Description object is created from a lab description file. The lab description file is a YAML file with the following structure.

        Parameters:
            path: Path to the lab edition file.
            platform: Platform to deploy the lab. Valid values are "aws" and "libvirt".
            lab_repo_uri: URI of the lab repository.
            teacher_access: Type of access to the teacher machine. Valid values are "host" and "instance".
            configure_dns: Whether to configure DNS or not.
            ssh_public_key_file: Path to the SSH public key file.
            ansible_ssh_common_args: Ansible SSH common arguments.
            aws_region: AWS region to deploy the lab.
            aws_default_instance_type: Default instance type for AWS.
            network_cidr_block: CIDR block for the lab network.
            packetbeat_policy_name: Name of the Packetbeat policy.
            packetbeat_vlan_id: VLAN ID for Packetbeat.
            elastic_stack_version: Elastic Stack version.
            libvirt_uri: URI for libvirt.
            libvirt_storage_pool: Storage pool for libvirt.
            libvirt_student_access: Type of access to the student machines. Valid values are "bridge" and "nat".
            libvirt_bridge: Name of the bridge for libvirt.
            libvirt_external_network: External network for libvirt.
            libvirt_bridge_base_ip: Base IP for the libvirt bridge.
            proxy: Proxy for libvirt.
            instance_type: An InstanceType object to compute the correct size of machine.
            endpoint_policy_name: Name of the Agent policy.
            user_install_packetbeat: User used to install Packetbeat.
            internet_network_cidr_block: CIDR of internet network.
            services_network_cidr_block: CIDR of services network.
            keep_ansible_logs: Keep Ansible logs on managed hosts.
            docker_uri: URI for docker.
            caldera_version: Caldera version.
            docker_dns: DNS to use for internet networks on Docker.
            ansible_forks: Number of parallel connections for Ansible.
            ansible_pipelining: Enable pipelining for Ansible.
            ansible_timeout: Timeout for Ansible connections.

        Returns:
             A Description object.
        """
        self.lab_edition_file = Path(path).resolve().as_posix()
        base_dir = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../.."))
        
        self.lab_repo_uri = lab_repo_uri
        if self.lab_repo_uri and not Path(self.lab_repo_uri).is_absolute():
            self.lab_repo_uri = Path(base_dir).joinpath(
                self.lab_repo_uri
            )
        self.platform = platform
        self.teacher_access = teacher_access
        self.configure_dns = configure_dns
        self.ansible_ssh_common_args = ansible_ssh_common_args
        self.aws_region = aws_region
        self.aws_default_instance_type = aws_default_instance_type
        self.network_cidr_block = network_cidr_block
        self.packetbeat_policy_name = packetbeat_policy_name
        self.packetbeat_vlan_id = packetbeat_vlan_id
        self.elastic_stack_version = elastic_stack_version
        self.is_elastic_stack_latest_version = False
        self.libvirt_uri = libvirt_uri
        self.libvirt_storage_pool = libvirt_storage_pool
        self.libvirt_student_access = libvirt_student_access
        self.libvirt_bridge = libvirt_bridge
        self.libvirt_external_network = libvirt_external_network
        self.libvirt_bridge_base_ip = libvirt_bridge_base_ip
        self.proxy = proxy
        self.instance_type = instance_type
        self.endpoint_policy_name = endpoint_policy_name
        self.user_install_packetbeat = user_install_packetbeat
        self.internet_network = internet_network_cidr_block
        self.services_network = services_network_cidr_block
        self.keep_ansible_logs = keep_ansible_logs
        self.docker_uri = docker_uri
        self.caldera_version = caldera_version
        self.docker_dns = docker_dns
        self.services = {}
        self._load_lab_edition(path)
        self.description_dir = Path(self.description_file).parent.resolve().as_posix()
        self.ansible_playbooks_path = (
            Path(self.description_dir).joinpath("ansible").resolve().as_posix()
        )

        self.advanced_options_path = Path(self.description_dir).joinpath("advanced").joinpath(platform).resolve().as_posix()

        self.ssh_public_key_file = ssh_public_key_file
        self.authorized_keys = self.read_pubkeys(
            self.teacher_pubkey_dir, ssh_public_key_file
        )
        self.ansible_forks = ansible_forks
        self.ansible_pipelining = ansible_pipelining
        self.ansible_timeout = ansible_timeout

    def read_pubkeys(self, ssh_dir, default_pubkey=None):
        """
        Reads all public keys in the ssh directory of the description folder.

        Parameters:
            ssh_dir (str): Directory to search for public keys.
            default_pubkey (str): Pubkey to use if there are no keys in ssh_dir.

        Returns:
            str: string with the contents of the ssh public keys.
        """
        keys = ""
        if ssh_dir and Path(ssh_dir).is_dir():
            for child in Path(ssh_dir).iterdir():
                if child.is_file():
                    key = child.read_text()
                    if key[-1] != "\n":
                        key += "\n"
                    keys += key

        if default_pubkey is not None:
            p = Path(default_pubkey).expanduser()
            if p.is_file():
                key = p.read_text()
                if key[-1] != "\n":
                    key += "\n"
                keys += key

        return keys

    def _required(self, description, field):
        """
        Check if the field is in the description.

        Parameters:
            description (obj): description.
            field (str): field to be searched.
        """
        if not description.get(field):
            raise DescriptionException(f"{field} not defined.")

    def _validate_value(self, field, value, values):
        if not value in values:
            raise DescriptionException(f'Error in {field} field. The possible values are "{values}" but "{value}" value was assigned.')
        
    def _validate_description(self, description):
        """
        Apply validations to lab description.

        Parameters:
            description (obj): lab description.
        """
        self._required(description, "institution")
        self._required(description, "lab_name")
        self._required(description, "guest_settings")
        self._required(description, "topology")

        self._validate_value("elastic_settings.enable", description.get("elastic_settings",{}).get("enable",False), [True, False])
        self._validate_value("elastic_settings.monitor_type", description.get("elastic_settings",{}).get("monitor_type","traffic"), ["traffic", "endpoint"])
        self._validate_value("caldera_settings.enable", description.get("caldera_settings",{}).get("enable",False), [True, False])

    def _validate_lab_edition(self, lab_edition_info):
        """
        Apply validations to lab edition specification.

        Parameters:
            lab_edition_info (obj): lab edition specification.
        """
        self._required(lab_edition_info, "base_lab")
        self._required(lab_edition_info, "instance_number")
        if lab_edition_info.get("create_student_passwords"):
            self._required(lab_edition_info, "random_seed")
        self._validate_value("elastic_settings.enable", lab_edition_info.get("elastic_settings",{}).get("enable",False), [True, False])
        self._validate_value("caldera_settings.enable", lab_edition_info.get("caldera_settings",{}).get("enable",False), [True, False])


    def _load_description(self, description_file):
        """
        Load lab description file.

        Parameters:
            description_file (str): path to lab description file.
        """
        stream = open(description_file, "r")
        description = yaml.safe_load(stream)
        self._validate_description(description)

        self.base_institution = re.sub("[^a-zA-Z0-9]+", "", description["institution"])
        self.base_lab_name = re.sub("[^a-zA-Z0-9]+", "", description["lab_name"])
        self.default_os = description.get("default_os", self.default_os)
        self.guest_settings = {key.lower(): value for key, value in description.get("guest_settings", self.guest_settings).items()}
        self.topology = description.get("topology", self.topology)

        self._load_elastic_settings(description)
        self._load_caldera_settings(description)

    def _load_lab_edition(self, path):
        """
        Load lab edition specification file.

        Parameters:
            path (str): path to lab edition specification file.
        """
        stream = open(path, "r")
        lab_edition_info = yaml.safe_load(stream)
        self._validate_lab_edition(lab_edition_info)

        self.instance_number = lab_edition_info["instance_number"]
        self.base_lab = lab_edition_info["base_lab"]

        self.student_prefix = lab_edition_info.get("student_prefix", "trainee")
        self.student_pubkey_dir = lab_edition_info.get("student_pubkey_dir")
        if self.student_pubkey_dir and not Path(self.student_pubkey_dir).is_absolute():
            self.student_pubkey_dir = (
                Path(self.lab_edition_file)
                .parent.joinpath(self.student_pubkey_dir)
                .resolve()
                .as_posix()
            )

        self.create_student_passwords = lab_edition_info.get(
            "create_student_passwords", False
        )
        self.random_seed = lab_edition_info.get("random_seed")

        # Open the lab directory if it exists, otherwise use a CTF package file
        if Path(self.lab_repo_uri).joinpath(self.base_lab).is_dir():
            self.description_file = Path(self.lab_repo_uri).joinpath(
                self.base_lab, "description.yml"
            )
            self.parameters_files = read_files_in_directory(Path(self.lab_repo_uri).joinpath(self.base_lab,"ansible","parameters"))
        else:
            lab_pkg_file = Path(self.lab_repo_uri).joinpath(f"{self.base_lab}.ctf")
            if Path(lab_pkg_file).is_file():
                pkg = ZipFile(lab_pkg_file)
                # Extract the package to temporary directory
                self.extract_tmpdir = tempfile.TemporaryDirectory(
                    prefix="tectonic", suffix=self.base_lab
                )
                pkg.extractall(path=self.extract_tmpdir.name)
                self.description_file = Path(self.extract_tmpdir.name).joinpath(
                    "description.yml"
                )
                self.parameters_files = read_files_in_directory(Path(self.extract_tmpdir.name,"ansible","parameters"))
            else:
                raise DescriptionException(f"{self.base_lab} not found in {self.lab_repo_uri}.")
        self._load_description(self.description_file)

        self.institution = str.lower(re.sub(
            "[^a-zA-Z0-9]+",
            "",
            lab_edition_info.get("institution", self.base_institution),
        ))
        self.lab_name = str.lower(re.sub(
            "[^a-zA-Z0-9]+",
            "",
            lab_edition_info.get("lab_edition_name", self.base_lab_name),
        ))

        self.teacher_pubkey_dir = lab_edition_info.get("teacher_pubkey_dir")
        if self.teacher_pubkey_dir and not Path(self.teacher_pubkey_dir).is_absolute():
            self.teacher_pubkey_dir = Path(self.lab_edition_file).parent.joinpath(
                self.teacher_pubkey_dir
            )

        self.deploy_elastic = self.deploy_elastic and lab_edition_info.get("elastic_settings",{}).get("enable", True)
        if self.deploy_elastic:
            self.services["elastic"] = {
                "vcpu": lab_edition_info.get("elastic_settings",{}).get("vcpu", 4),
                "memory": lab_edition_info.get("elastic_settings",{}).get("memory", 8192),
                "disk": lab_edition_info.get("elastic_settings",{}).get("disk", 50)
            }
            if self.platform == "aws":
                self.services["packetbeat"] = {
                    "vcpu": 1,
                    "memory": 512,
                    "disk": 10,
                }
        self.deploy_caldera = self.deploy_caldera and lab_edition_info.get("caldera_settings",{}).get("enable", True)
        if self.deploy_caldera:
            self.services["caldera"] = {
                "vcpu": lab_edition_info.get("caldera_settings",{}).get("vcpu", 2),
                "memory": lab_edition_info.get("caldera_settings",{}).get("memory", 2048),
                "disk": lab_edition_info.get("caldera_settings",{}).get("disk", 20)
            }

        self._expand_topology()
        self.subnets = self._compute_subnetworks()

    def _load_elastic_settings(self, description):
        settings = description.get("elastic_settings", {})
        self.deploy_elastic = settings.get("enable",False)
        self.monitor_type = settings.get("monitor_type","traffic")
        self.elastic_deploy_default_policy = settings.get("deploy_default_policy",True)

    def _load_caldera_settings(self, description):
        settings = description.get("caldera_settings", {})
        self.deploy_caldera = settings.get("enable",False)

    def get_instance_range(self):
        """
        Get range for instances.

        Returns:
            range: instance range.
        """
        return range(1, self.instance_number + 1)

    def get_guest_attr(self, guest_name, attr, default=None):
        """
        Get guest specific attribute

        Parameters:
            guest_name (str): machine guest name.
            attr (str): attribute to return.
            default (str, None): default value to return in case attribute is not found.

        Returns:
            any: attribute value for the guest.
        """
        if self.guest_settings.get(guest_name):
            return self.guest_settings[guest_name].get(attr, default)
        else:
            return default

    def get_guest_copies(self, guest_name):
        """
        Get guest number of copies

        Parameters:
            guest_name (str): machine guest name.

        Returns:
            int: copies number.
        """
        return self.get_guest_attr(guest_name, "copies", 1)

    def get_copy_range(self, guest_name):
        """
        Get range for guest instance.

        Parameters:
            guest_name (str): machine guest name.

        Returns:
            range: copy range.
        """
        return range(1, self.get_guest_copies(guest_name) + 1)

    def get_image_name(self, guest_name):
        """
        Get image full name.

        Parameters:
            guest_name (str): machine guest name.

        Returns:
            str: full name.
        """
        return f"{self.institution}-{self.lab_name}-{guest_name}"

    def get_instance_name(self, instance_num, guest_name, copy=1):
        """
        Get machine full name.

        Parameters:
            instance_num (int): machine instance number.
            guest_name (str): machine guest name.
            copy (int): machine copy number.

        Returns:
            str: full name.
        """
        return f"{self.institution}-{self.lab_name}-{instance_num}-{guest_name}" + (
            ("-" + str(copy)) if self.get_guest_copies(guest_name) > 1 else ""
        )

    def get_hostname(self, instance_num, guest_name, copy=1):
        """
        Get hostname.

        Parameters:
            instance_num (int): machine instance number.
            guest_name (str): machine guest name.
            copy (int): machine copy number.

        Returns:
            str: hostname.
        """
        return f"{guest_name}-{instance_num}" + (
            ("-" + str(copy)) if self.get_guest_copies(guest_name) > 1 else ""
        )

    def get_instance_number(self, instance_name):
        """
        Get instance number.

        Parameters:
            instance_name (str): full name of the instance.

        Returns:
            int: instance number.
        """
        tokens = instance_name.split("-")
        if len(tokens) >= 3 and tokens[2].isdigit():
            return int(tokens[2])
        else:
            return None

    def get_base_name(self, instance_name):
        """
        Get instance base name (guest).

        Parameters:
            instance_name (str): full name of the instance.

        Returns:
            str: instance guest name.
        """
        tokens = instance_name.split("-")
        if len(tokens) == 3:
            return tokens[2]
        elif len(tokens) >= 4:
            return tokens[3]
        else:
            return None

    def get_copy(self, instance_name):
        """
        Get instance copy.

        Parameters:
            instance_name (str): full name of the instance.

        Returns:
            int: instance copy number.
        """
        tokens = instance_name.split("-")
        if len(tokens) >= 5 and tokens[4].isdigit():
            return int(tokens[4])
        else:
            return 1

    def get_student_access_name(self):
        """
        Returns student access machine full name.

        Returns:
            str: student access machine name.
        """
        return f"{self.institution}-{self.lab_name}-student_access"

    def get_teacher_access_name(self):
        """
        Returns teacher access machine full name.

        Returns:
            str: teacher access machine name.
        """
        return f"{self.institution}-{self.lab_name}-teacher_access"

    def get_service_name(self, service):
        """
        Returns service machine full name.

        Returns:
            str: service machine name.
        """
        return f"{self.institution}-{self.lab_name}-{service}"

    def get_services_to_deploy(self):
        "Return services name to be deploy"
        services = []
        if self.deploy_elastic:
            services.append("elastic")
            if self.monitor_type == "traffic" and self.platform == "aws":
                services.append("packetbeat")
        if self.deploy_caldera:
            services.append("caldera")
        return services
    
    def parse_machines(self, instances=None, guests=(), copies=None, only_instances=True, exclude=[]):
        """
        Return machines names based on instance number, guest name and number of copy.

        Parameters:
            instances (list(int)): numbers of instances.
            guests (tuple(str)): guest name of machines.
            copies (list(int)): number of copy of the machines.
            only_instances (bool): if true return only scenario machines. Otherwise also return aux machines.
            exclude (list(str)): full name of machines to exclude.

        Returns:
            list(str): full name of machines.
        """
        if instances is None:
            instances = []
        infrastructure_guests_names = []
        if self.platform == "aws":
            infrastructure_guests_names = []
            if self.is_student_access():
                infrastructure_guests_names.append("student_access")
            if self.teacher_access == "host":
                infrastructure_guests_names.append("teacher_access")
        for service in self.get_services_to_deploy():
            infrastructure_guests_names.append(service)

        if guests is None:
            guests = ()
        guests_aux = list(self.guest_settings.keys())
        if not only_instances:
            guests_aux = guests_aux + infrastructure_guests_names
        if instances and max(instances) > self.instance_number:
            raise DescriptionException("Invalid instance number specified.")
        if not set(guests).issubset(set(guests_aux)):
            raise DescriptionException("Invalid guests names specified.")
        max_guest_copy = 0
        for guest in guests or guests_aux:
            guest_copies = self.get_guest_attr(guest, "copies", 0)
            if guest_copies > max_guest_copy:
                max_guest_copy = guest_copies
        if copies and max_guest_copy > 0 and max(copies) > max_guest_copy:
            raise DescriptionException("Invalid copies specified.")

        if max_guest_copy == 0 and copies and copies != [1]:
            # The user specified a copy number other than one, but
            # no guest has copies defined.
            raise DescriptionException(
                "No machines with the specified characteristics were found."
            )

        result = []
        for instance_num in self.get_instance_range():
            for guest_name in self.guest_settings.keys():
                for copy in self.get_copy_range(guest_name):
                    result.append(self.get_instance_name(instance_num, guest_name, copy))

        if not only_instances :
            if self.platform == "aws":
                result.append(self.get_student_access_name())
                if self.teacher_access == "host":
                    result.append(self.get_teacher_access_name())

            services = self.get_services_to_deploy()
            for service in services:
                result.append(self.get_service_name(service))

        if instances is not None and instances != []:
            instance_re = f"^{self.institution}-{self.lab_name}-({'|'.join(str(instance) for instance in instances)})-"
            result = list(
                filter(
                    lambda machine: re.search(instance_re, machine) is not None, result
                )
            )

        if guests:
            guest_re = (
                rf"^{self.institution}-{self.lab_name}(-\d+)?-({'|'.join(guests)})(-|$)"
            )
            result = list(
                filter(lambda machine: re.search(guest_re, machine) is not None, result)
            )

        if copies:
            one_copy_re = rf"^{self.institution}-{self.lab_name}-\d+-[^-]+$"
            many_copies_re = rf"^{self.institution}-{self.lab_name}-\d+-[^-]+-({'|'.join(str(copy) for copy in copies)})$"
            result = list(
                filter(
                    lambda machine: re.search(
                        many_copies_re
                        if (self.get_guest_copies(self.get_base_name(machine)) > 1)
                        or (1 not in copies)
                        else one_copy_re,
                        machine,
                    )
                    is not None,
                    result,
                )
            )

        for machine in exclude:
            guest_re = (
                rf"^{self.institution}-{self.lab_name}(-\d+)?-{machine}(-|$)"
            )
            result = list(
                filter(lambda machine: re.search(guest_re, machine) is None, result)
            )

        if len(result) == 0:
            raise DescriptionException(
                "No machines with the specified characteristics were found."
            )
        else:
            return result

    def get_machines_to_monitor(self):
        """Returns the list of machines that have the monitor attribute set."""
        monitor = []
        for guest_name, guest in self.guest_settings.items():
            if guest and guest.get("monitor"):
                monitor.append(guest_name)
        return monitor

    def get_red_team_machines(self):
        """Returns the list of machines that have the red_team_agent set."""
        red_team = []
        for guest_name, guest in self.guest_settings.items():
            if guest and guest.get("red_team_agent", False):
                red_team.append(guest_name)
        return red_team

    def get_blue_team_machines(self):
        """Returns the list of machines that have the blue_team_agent set."""
        blue_team = []
        for guest_name, guest in self.guest_settings.items():
            if guest and guest.get("blue_team_agent", False):
                blue_team.append(guest_name)
        return blue_team

    def get_guest_username(self, guest_name):
        """
        Returns username to connect to guest.

        Parameters:
            guest_name (str): guest name.

        Returns:
            str: username
        """
        if guest_name in self.get_services_to_deploy():
            base_os = self.get_service_base_os(guest_name)
        else:
            base_os = self.get_guest_attr(guest_name, "base_os", self.default_os)
        os_data = OS_DATA.get(base_os)
        if not os_data:
            raise DescriptionException(f"Invalid operating system {base_os}.")

        return os_data["username"]

    def get_guest_networks(self, guest_name):
        """
        Returns the name of all the networks the guest is connected to.

        Parameters:
            guest_name (str): name of the guest

        Returns:
            list(str): networks for the guest
        """
        networks = []
        for network in self.topology:
            if guest_name in network["members"]:
                networks.append(network["name"])
        return networks

    def is_internet_access(self):
        """
        Return if internet access is required

        Returns:
            bool: True if internet access is required for any guest; False otherwhise.
        """
        for guest in self.guest_settings.values():
            if guest and guest.get("internet_access"):
                return True
        return False
    
    def is_student_access(self):
        """
        Return if student access is required

        Returns:
            bool: True if student access is required for any guest; False otherwhise.
        """
        for guest in self.guest_settings.values():
            if guest and guest.get("entry_point"):
                return True
        return False

    def get_subnetwork_name(self, instance_num, network_name):
        """Returns the name of the instance subnetwork."""
        return f"{self.institution}-{self.lab_name}-{instance_num}-{network_name}"

    def _compute_subnetworks(self):
        """Compute the complete list of subnetworks."""
        subnets = {}

        new_subnet_bits = math.ceil(math.log2(len(self.topology)))
        for instance_num in self.get_instance_range():
            instance_network = list(
                ipaddress.ip_network(self.network_cidr_block).subnets(prefixlen_diff=8)
            )[instance_num]
            instance_subnets = list(instance_network.subnets(new_subnet_bits))
            for i, n in enumerate(self.topology):
                subnet_name = self.get_subnetwork_name(instance_num, n["name"])
                subnets[subnet_name] = str(instance_subnets[i])

        return subnets

    def _expand_topology(self):
        """Expand network members with copies"""
        for n in self.topology:
            n["members"] = [
                member
                for member in n["members"]
                for copy in range(self.get_guest_copies(member))
            ]

    def get_topology_network(self, network_name):
        """Returns the topology information of the network."""
        for n in self.topology:
            if n["name"] == network_name:
                return n
        return None

    def _get_ipv4_subnet(self, instance_num, network_name):
        """Returns an IPv4Network object for an instance network."""
        return ipaddress.ip_network(
            self.subnets[self.get_subnetwork_name(instance_num, network_name)]
        )

    def get_instance_ip_address(self, instance_num, guest_name, copy, network):
        """Compute the IP address of the given instance in the network."""
        hostnum = network["members"].index(guest_name) + (copy - 1) + 3
        return str(
            list(self._get_ipv4_subnet(instance_num, network["name"]).hosts())[hostnum]
        )

    def get_guest_data(self):
        """Compute the guest data as expected by the deployment terraform module."""

        guest_data = {}
        entry_point_index = 1
        services_network_index = 1
        for instance_num in self.get_instance_range():
            for base_name, guest in self.guest_settings.items():
                for copy in self.get_copy_range(base_name):
                    guest_name = self.get_instance_name(instance_num, base_name, copy)
                    interfaces = {}
                    for index, network in enumerate([n for n in self.topology if base_name in n["members"]]):
                        private_ip = self.get_instance_ip_address(instance_num, base_name, copy, network)
                        interface = {
                            "name": f"{guest_name}-{index + 1}",
                            "index": index + self._get_interface_index_to_sum(base_name),
                            "guest_name": guest_name,
                            "network_name": network["name"],
                            "subnetwork_name": f"{self.institution}-{self.lab_name}-{instance_num}-{network['name']}",
                            "private_ip": private_ip,
                            "mask": self._get_ipv4_subnet(instance_num, network["name"]).prefixlen,
                        }
                        interfaces[interface["name"]] = interface

                    memory = self.get_guest_attr(base_name, "memory", 1024)
                    vcpus = self.get_guest_attr(base_name, "vcpu", 1)
                    gpu = self.get_guest_attr(base_name, "gpu", False)
                    monitor = self.get_guest_attr(base_name, "monitor", False)
                    is_in_services_network = self.is_in_services_network(base_name)
                    is_entry_point = self.get_guest_attr(base_name, "entry_point", False)
                    guest_data[guest_name] = {
                        "guest_name": guest_name,
                        "base_name": base_name,
                        "instance": instance_num,
                        "copy": copy,
                        "hostname": self.get_hostname(instance_num, base_name, copy),
                        "entry_point": is_entry_point,
                        "entry_point_index": entry_point_index,
                        "internet_access": self.get_guest_attr(base_name, "internet_access", False),
                        "base_os": self.get_guest_attr(base_name, "base_os", self.default_os),
                        "interfaces": interfaces,
                        "memory": memory,
                        "vcpu": vcpus,
                        "disk": self.get_guest_attr(base_name, "disk", 10),
                        "instance_type": self.instance_type.get_guest_instance_type(memory,
                                                                                    vcpus,
                                                                                    gpu,
                                                                                    monitor,
                                                                                    self.monitor_type
                                                                                    ),
                        "advanced_options_file": self._get_guest_advanced_options_file(base_name),
                        "is_in_services_network" : is_in_services_network,
                        "services_network_index": services_network_index
                    }
                    if is_entry_point:
                        entry_point_index += 1
                    if is_in_services_network:
                        services_network_index += 1

        return guest_data
    
    def is_in_services_network(self, base_name):
        return (
            (self.deploy_elastic and self.monitor_type == "endpoint" and self.get_guest_attr(base_name, "monitor", False)) or 
            (self.deploy_caldera and (self.get_guest_attr(base_name, "red_team_agent", False) or self.get_guest_attr(base_name, "blue_team_agent", False)))
        )

    def _get_interface_index_to_sum(self, base_name):
        """
        Returns number to be added to interface index
        """
        is_in_services_network = self.is_in_services_network(base_name)
        if self.platform == "libvirt":
            base = 3
            if self.get_guest_attr(base_name, "entry_point", False):
                if is_in_services_network:
                    return base + 2
                else:
                    return base + 1
            else:
                if is_in_services_network:
                    return base + 1
                else:
                    return base
        elif self.platform == "aws" or self.platform == "docker":
            return 0


    def set_elastic_stack_version(self, version):
        """
        Set Elastic Stack version

        Parameters:
            version (str): Elastic Stack version
        """
        self.elastic_stack_version = version
        self.is_elastic_stack_latest_version = True

    def set_caldera_version(self, version):
        """
        Set Caldera version

        Parameters:
            version (str): Caldera version
        """
        self.caldera_version = version

    def _get_guest_advanced_options_file(self, base_name):
        """
        Return path to advanced options for the guest if exists or /dev/null otherwise.

        Parameters:
            base_name (str): Machine base name
        """
        if self.platform == "libvirt":
            if (Path(self.advanced_options_path).joinpath(f'{base_name}.xsl')).is_file():
                return Path(self.advanced_options_path).joinpath(f'{base_name}.xsl').resolve().as_posix()
            else:
                return "/dev/null" 
        else:
                return "/dev/null"
    
    def get_parameters(self, instances=None):
        if not instances:
            instances = list(range(1,self.instance_number+1))
        random.seed(self.random_seed)
        choices = {}
        for file in self.parameters_files:
            choices[Path(file).stem] = random.choices(open(file).readlines(),k=self.instance_number)
        parameters = {}
        for instance in instances:
            parameter = {}
            for choice in choices:
                parameter[choice] = json.loads(choices[choice][instance-1])
            parameters[instance] = parameter
        return parameters

    def get_service_base_os(self, service_name):
        if service_name == "elastic" or service_name == "caldera":
            return "rocky8"
        elif service_name == "packetbeat":
            return "ubuntu22"
        




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import json
import click
import ipaddress
import datetime
import time
import math

from tectonic.deployment import Deployment, DeploymentException
from tectonic.docker_client import Client
from tectonic.constants import OS_DATA
from tectonic.utils import create_table
from tectonic.ansible import Ansible

import importlib.resources as tectonic_resources



class DeploymentDockerException(DeploymentException):
    pass

class DockerDeployment(Deployment):

    def __init__(
        self,
        description,
        gitlab_backend_url,
        gitlab_backend_username,
        gitlab_backend_access_token,
        packer_executable_path="/usr/bin/packer",
    ):
        super().__init__(
            description,
            Client(description),
            gitlab_backend_url,
            gitlab_backend_username,
            gitlab_backend_access_token,
            packer_executable_path
        )

    def get_deploy_cr_vars(self):
        """Build the terraform variable dictionary for deployment of the CyberRange."""
        return {
            "institution": self.description.institution,
            "lab_name": self.description.lab_name,
            "instance_number": self.description.instance_number,
            "ssh_public_key_file": self.description.ssh_public_key_file,
            "authorized_keys": self.description.authorized_keys,
            "subnets_json": json.dumps(self.description.subnets),
            "guest_data_json": json.dumps(self.description.get_guest_data()),
            "default_os": self.description.default_os,
            "os_data_json": json.dumps(OS_DATA),
            "configure_dns": self.description.configure_dns,
            "docker_uri": self.description.docker_uri,
            }
    
    def get_deploy_services_vars(self):
        """Build the terraform variable dictionary for deployment of the CyberRange services."""
        return {
            "institution": self.description.institution,
            "lab_name": self.description.lab_name,
            "ssh_public_key_file": self.description.ssh_public_key_file,
            "authorized_keys": self.description.authorized_keys,
            "subnets_json": json.dumps(self._get_services_network_data()),
            "guest_data_json": json.dumps(self._get_services_guest_data()),
            "os_data_json": json.dumps(OS_DATA),
            "configure_dns": self.description.configure_dns,
            "docker_uri": self.description.docker_uri
        }

    def delete_cr_images(self, guests=None):
        """
        Delete guests base image.
        """
        guests = guests or self.description.guest_settings.keys()
        for guest_name in guests:
            try: 
                self.client.delete_image(self.description.get_image_name(guest_name))
            except Exception as exception:
                raise DeploymentDockerException(f"{exception}")
            
    def delete_services_images(self, services):
        """
        Delete services base image.
        """
        for service in services:
            if services[service]:
                try: 
                    self.client.delete_image(service)
                except Exception as exception:
                    raise DeploymentDockerException(f"{exception}")
            
    def create_cr_images(self, guests=None):
        super().create_cr_images(guests)

    def create_services_images(self, services):
        super().create_services_images(services) 

    def deploy_infraestructure(self, instances):
        """
        Deploy cyber range infraestructure
        """
        ansible = Ansible(self)

        if len(self.description.get_services_to_deploy()) > 0: #Deploy services
            click.echo("Deploying Cyber Range services...")
            self._deploy_services(
                tectonic_resources.files('tectonic') / 'services' / 'terraform' / 'services-docker',
                self.get_deploy_services_vars(),
                instances
            )

            click.echo("Waiting for services to boot up...")
            services_to_deploy = []
            for service in self.description.get_services_to_deploy():
                services_to_deploy.append(self.description.get_service_name(service))
            extra_vars = {
                "elastic" : {
                    "monitor_type": self.description.monitor_type,
                    "deploy_policy": self.description.elastic_deploy_default_policy,
                    "policy_name": self.description.packetbeat_policy_name if self.description.monitor_type == "traffic" else self.description.endpoint_policy_name,
                    "http_proxy" : self.description.proxy,
                    "description_path": self.description.description_dir,
                    "ip": str(ipaddress.IPv4Network(self.description.services_network)[2]),
                    "elasticsearch_memory": math.floor(self.description.services["elastic"]["memory"] / 1000 / 2)  if self.description.deploy_elastic else None,
                    "dns": self.description.docker_dns,
                },
                "caldera":{
                    "ip": str(ipaddress.IPv4Network(self.description.services_network)[4]),
                    "description_path": self.description.description_dir,
                },
            }
            inventory = ansible.build_inventory(machine_list=services_to_deploy, extra_vars=extra_vars)
            ansible.wait_for_connections(inventory=inventory)

            click.echo("Configuring services...")
            ansible.run(
                inventory=inventory,
                playbook=self.ansible_services_path,
                quiet=True,
            )
            if self.description.deploy_elastic and self.description.monitor_type == "traffic":
                self._deploy_packetbeat()

        click.echo("Deploying Cyber Range instances...")
        self._deploy_cr(tectonic_resources.files('tectonic') / 'terraform' / 'modules' / 'gsi-lab-docker',
                        self.get_deploy_cr_vars(),
                        instances)
        
        click.echo("Waiting for machines to boot up...")
        ansible.wait_for_connections(instances=instances)

        click.echo("Configuring student access...")
        self.student_access(instances)

        click.echo("Running after-clone configuration...")
        ansible.run(instances, quiet=True)

        if self.description.deploy_elastic and self.description.monitor_type == "endpoint":
            click.echo("Configuring elastic agents...")
            self._elastic_install_endpoint(instances)
        
        if self.description.deploy_caldera:
            click.echo("Configuring caldera agents...")
            self._caldera_install_agent(instances)

    def destroy_infraestructure(self, instances):
        """
        Destroy cyber range infraestructure
        """
        click.echo("Destroying Cyber Range instances...")
        self._destroy_cr(
            tectonic_resources.files('tectonic') / 'terraform' / 'modules' / 'gsi-lab-docker',
            self.get_deploy_cr_vars(),
            instances
        )

        if instances is None and len(self.description.get_services_to_deploy()) > 0:
            click.echo("Destroying Cyber Range services...")
            self._destroy_services(
                tectonic_resources.files('tectonic') / 'services' / 'terraform' / 'services-docker',
                self.get_deploy_services_vars(),
                instances
            )
            if self.description.deploy_elastic and self.description.monitor_type == "traffic":
                self._destroy_packetbeat()
    
    def get_cyberrange_data(self):
        """Get information about cyber range"""
        try:
            if len(self.description.get_services_to_deploy()) > 0:
                headers = ["Name", "Value"]
                rows = []
                if self.description.deploy_elastic:
                    elastic_name = self.description.get_service_name("elastic")
                    if self.get_instance_status(elastic_name) == "RUNNING":
                        elastic_credentials = self._get_service_password("elastic")
                        #elastic_ip = self.get_ssh_hostname(elastic_name)
                        elastic_ip = "127.0.0.1"
                        rows.append(["Kibana URL", f"https://{elastic_ip}:5601"])
                        rows.append(["Kibana user (username: password)", f"elastic: {elastic_credentials['elastic']}"])
                        if self.description.deploy_caldera:
                            rows.append(["",""])
                    else:
                        return "Unable to get Elastic info right now. Please make sure de Elastic machine is running."
                if self.description.deploy_caldera:
                    caldera_name = self.description.get_service_name("caldera")
                    if self.get_instance_status(caldera_name) == "RUNNING":
                        caldera_credentials = self._get_service_password("caldera")
                        #caldera_ip = self.get_ssh_hostname(caldera_name)
                        caldera_ip = "127.0.0.1"
                        rows.append(["Caldera URL", f"https://{caldera_ip}:8443"])
                        rows.append(["Caldera user (username: password)", f"red: {caldera_credentials['red']}"])
                        rows.append(["Caldera user (username: password)", f"blue: {caldera_credentials['blue']}"])
                    else:
                        return "Unable to get Caldera info right now. Please make sure de Caldera machine is running."
                return create_table(headers,rows)
        except Exception as exception:
            raise DeploymentDockerException(f"{exception}")
        
    def get_instance_status(self, instance_name):
        """Returns a dictionary with the instance status."""
        return self.client.get_instance_status(instance_name)
        
    def list_instances(self, instances, guests, copies):
        machines_to_list = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())
        headers = ["Name", "Status"]
        rows = []
        for machine in machines_to_list:
            machine_status = self.get_instance_status(machine)
            rows.append([machine,machine_status])
        return create_table(headers,rows)
            
    def start_instance(self, instance_name):
        self.client.start_instance(instance_name)

    def stop_instance(self, instance_name):
        self.client.stop_instance(instance_name)

    def reboot_instance(self, instance_name):
        self.client.reboot_instance(instance_name)

    def shutdown(self, instances, guests, copies, stop_services):
        if stop_services:
            click.echo(f"Shutting down services...")
            for service in self.description.get_services_to_deploy():
                self.stop_instance(self.description.get_service_name(service))
                if service == "elastic" and self.description.monitor_type == "traffic":
                    #TODO
                    pass             
        machines_to_shutdown = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy()) 
        for machine in machines_to_shutdown:
            click.echo(f"Shutting down instance {machine}...")
            self.stop_instance(machine)
            
    def start(self, instances, guests, copies, start_services):
        if start_services:
            click.echo(f"Booting up services...")
            for service in self.description.get_services_to_deploy():
                self.start_instance(self.description.get_service_name(service))
                if service == "elastic" and self.description.monitor_type == "traffic":
                    #TODO
                    pass
        machines_to_start = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())  
        for machine in machines_to_start:
            click.echo(f"Booting up instance {machine}...")
            self.start_instance(machine)

    def reboot(self, instances, guests, copies, reboot_services):
        if reboot_services:
            click.echo(f"Rebooting services...")
            for service in self.description.get_services_to_deploy():
                self.reboot_instance(self.description.get_service_name(service))
                if service == "elastic" and self.description.monitor_type == "traffic":
                    #TODO
                    pass
        machines_to_reboot = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())
        for machine in machines_to_reboot:
            click.echo(f"Rebooting instance {machine}...")
            self.reboot_instance(machine)

    def connect_to_instance(self, instance_name, username):
        username = username or self.description.get_guest_username(self.description.get_base_name(instance_name))
        self.client.connect(instance_name, username)

    def _get_services_network_data(self):
        """Compute the complete list of services subnetworks."""
        networks = {
            f"{self.description.institution}-{self.description.lab_name}-services" : {
                "cidr" : self.description.services_network,
                "mode": "none"
            },
        }
        if self.description.deploy_elastic or self.description.deploy_caldera :
            networks[f"{self.description.institution}-{self.description.lab_name}-internet"] = {
                "cidr" : self.description.internet_network,
                "mode" : "nat",
            }
        return networks
    
    def _get_services_guest_data(self):
        """Compute the services guest data as expected by the deployment terraform module."""
        guest_data = {}
        if self.description.deploy_elastic:
            guest_data[self.description.get_service_name("elastic")] = {
                    "guest_name": self.description.get_service_name("elastic"),
                    "base_name": "elastic",
                    "hostname": "elastic",
                    "base_os": self.description.get_service_base_os("elastic"),
                    "interfaces": {
                        f'{self.description.get_service_name("elastic")}-1' : {
                            "name": f'{self.description.get_service_name("elastic")}-1',
                            "guest_name": self.description.get_service_name("elastic"),
                            "network_name": "services",
                            "subnetwork_name": f"{self.description.institution}-{self.description.lab_name}-services",
                            "private_ip": str(ipaddress.IPv4Network(self.description.services_network)[2]),
                            "mask": str(ipaddress.ip_network(self.description.services_network).prefixlen),
                        },
                        f'{self.description.get_service_name("elastic")}-2' : {
                            "name": f'{self.description.get_service_name("elastic")}-2',
                            "guest_name": self.description.get_service_name("elastic"),
                            "network_name": "internet",
                            "subnetwork_name": f"{self.description.institution}-{self.description.lab_name}-internet",
                            "private_ip": str(ipaddress.IPv4Network(self.description.internet_network)[2]),
                            "mask": str(ipaddress.ip_network(self.description.internet_network).prefixlen),
                        },
                    },
                    "memory": self.description.services["elastic"]["memory"],
                    "vcpu": self.description.services["elastic"]["vcpu"],
                    "disk": self.description.services["elastic"]["disk"],
                    "port": 5601,
                }
        if self.description.deploy_caldera:
            guest_data[self.description.get_service_name("caldera")] = {
                    "guest_name": self.description.get_service_name("caldera"),
                    "base_name": "caldera",
                    "hostname": "caldera",
                    "base_os": self.description.get_service_base_os("caldera"),
                    "interfaces": {
                        f'{self.description.get_service_name("caldera")}-1' : {
                            "name": f'{self.description.get_service_name("caldera")}-1',
                            "guest_name": self.description.get_service_name("caldera"),
                            "network_name": "services",
                            "subnetwork_name": f"{self.description.institution}-{self.description.lab_name}-services",
                            "private_ip": str(ipaddress.IPv4Network(self.description.services_network)[4]),
                            "mask": str(ipaddress.ip_network(self.description.services_network).prefixlen),
                        },
                        f'{self.description.get_service_name("caldera")}-2' : {
                            "name": f'{self.description.get_service_name("caldera")}-2',
                            "guest_name": self.description.get_service_name("caldera"),
                            "network_name": "internet",
                            "subnetwork_name": f"{self.description.institution}-{self.description.lab_name}-internet",
                            "private_ip": str(ipaddress.IPv4Network(self.description.internet_network)[4]),
                            "mask": str(ipaddress.ip_network(self.description.internet_network).prefixlen),
                        },
                    },
                    "memory": self.description.services["caldera"]["memory"],
                    "vcpu": self.description.services["caldera"]["vcpu"],
                    "disk": self.description.services["caldera"]["disk"],
                    "port": 8443,
                }
        return guest_data
    
    def get_services_status(self):
        if len(self.description.get_services_to_deploy()) > 0:
            headers = ["Name", "Status"]
            rows = []
            if self.description.deploy_elastic:
                elastic_name = self.description.get_service_name("elastic")
                elastic_status = self.get_instance_status(elastic_name)
                rows = [[elastic_name, elastic_status]]
                if self.description.monitor_type == "traffic":
                    rows.append([self.description.get_service_name("packetbeat"), self._manage_packetbeat("status")])
                else:
                    try:
                        if elastic_status == "RUNNING":
                            playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'get_info.yml'
                            result = self._get_service_info("elastic",playbook,{"action":"agents_status"})
                            agents_status = result[0]['agents_status']
                            for key in agents_status:
                                rows.append([f"elastic-agents-{key}", agents_status[key]])
                        else:
                            click.echo(f"Unable to connect to Elastic. Check if machine is running.")
                    except Exception as e:
                        raise DeploymentDockerException(e)
                if self.description.deploy_caldera:
                    rows.append(["",""])
            if self.description.deploy_caldera:
                caldera_name = self.description.get_service_name("caldera")
                rows.append([caldera_name, self.get_instance_status(caldera_name)])
                try:
                    if self.get_instance_status(caldera_name) == "RUNNING":
                        playbook = tectonic_resources.files('tectonic') / 'services' / 'caldera' / 'get_info.yml'
                        result = self._get_service_info("caldera",playbook,{"action":"agents_status"})
                        response = result[0]['agents_status']
                        agents_status = {"alive": 0, "dead": 0, "pending_kill":0}
                        if len(response) > 0:
                            for agent in response: #TODO: see what the response is like when there are a large number of agents. pagination?
                                #Caldera uses this logic to define the state of the agent
                                now = int(time.time() * 1000) #Milliseconds since epoch
                                agent_last_seen = int((datetime.datetime.strptime(agent["last_seen"],"%Y-%m-%dT%H:%M:%SZ") - datetime.datetime(1970, 1, 1)).total_seconds() * 1000)
                                difference = now - agent_last_seen
                                if (difference <= 60000 and agent["sleep_min"] == 3 and agent["sleep_max"] == 3 and agent["watchdog"] == 1):
                                    agents_status["pending_kill"] = agents_status["pending_kill"] + 1
                                elif (difference <= 60000 or difference <= (agent["sleep_max"] * 1000)):
                                    agents_status["alive"] = agents_status["alive"] + 1
                                else:
                                    agents_status["dead"] = agents_status["dead"] + 1
                        for key in agents_status:
                            rows.append([f"caldera-agents-{key}", agents_status[key]])
                    else:
                        click.echo(f"Unable to connect to Caldera. Check if machine is running.")
                except Exception as e:
                    raise DeploymentDockerException(e)
            return create_table(headers,rows)
        else:
            return "No services were deployed."
        
    def recreate(self, instances, guests, copies, recreate_services=False):
        machines = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())
        resources_to_recreate = self.get_resources_to_recreate(instances, guests, copies)
        click.echo("Recreating machines...")
        self.terraform_recreate(tectonic_resources.files('tectonic') / 'terraform' / 'modules' / "gsi-lab-docker", resources_to_recreate)

        click.echo("Waiting for machines to boot up...")
        ansible = Ansible(self)
        ansible.wait_for_connections(instances, guests, copies, False, self.description.get_services_to_deploy())

        click.echo("Configuring student access...")
        entry_points = []
        for m in machines:
            base_name = self.description.get_base_name(m)
            if (base_name not in entry_points and (self.description.get_guest_attr(base_name, "entry_point", False) or base_name == "student_access")):
                entry_points.append(base_name)
        self._student_access(instances, entry_points)

        click.echo("Running after-clone configuration...")
        ansible.run(instances, guests, copies, quiet=True, only_instances=False)

        if self.description.deploy_elastic and self.description.monitor_type == "endpoint":
            click.echo("Configuring elastic agents...")
            self._elastic_install_endpoint(instances)

        if self.description.deploy_caldera:
            click.echo("Configuring caldera agents...")
            self._caldera_install_agent(instances)

    def get_resources_to_recreate(self, instances, guests, copies):
        machines_to_recreate = self.description.parse_machines(instances, guests, copies, True)
        resource_to_recreate = []
        for machine in machines_to_recreate:
            resource_to_recreate.append('docker_container.machines["' f"{machine}" '"]')
        return resource_to_recreate
    
    def get_cr_resources_to_target_apply(self, instances):
        resources = self._get_machines_resources_name(instances)
        resources = resources + self._get_subnet_resources_name(instances)
        return resources

    def get_cr_resources_to_target_destroy(self, instances):
        machines = self.description.parse_machines(instances, None, None, True)
        resources = []
        for machine in machines:
            resources.append('docker_container.machines["' f"{machine}" '"]')
        return resources
    
    def _get_machines_resources_name(self, instances):
        machines = self.description.parse_machines(instances, None, None, True)
        resources = []
        for machine in machines:
            resources.append('docker_container.machines["' f"{machine}" '"]')
        return resources

    def _get_subnet_resources_name(self, instances):
        resources = []
        for instance in filter(
            lambda i: i <= self.description.instance_number, instances or range(1, self.description.instance_number+1)
        ):
            for network in self.description.topology:
                resources.append(
                    'docker_network.subnets["'
                    f"{self.description.institution}-{self.description.lab_name}-{str(instance)}-{network['name']}"
                    '"]'
                )
        return resources
    
    def get_services_resources_to_target_destroy(self, instances):
        """
        Returns the name of the libvirt resource of the services-libvirt module to target destroy base on the instances number

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name to target destroy
        """
        return []
        
    def get_services_resources_to_target_apply(self, instances):
        """
        Returns the name of the libvirt resource of the services-libvirt module to target apply base on the instances number

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name to target apply
        """
        resources = []
        for service in self._get_services_guest_data():
            resources.append('docker_container.machines["'f"{service}"'"]')
            resources.append('docker_image.base_images["'f"{service}"'"]')
        for network in self._get_services_network_data():
            resources.append('docker_network.subnets["'f"{network}"'"]')
        return resources
    
    def _deploy_packetbeat(self):
        try:
            elastic_name = self.description.get_service_name("elastic")
            if self.get_instance_status(elastic_name) == "RUNNING":
                elastic_ip = self.get_ssh_hostname(elastic_name)
                playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'get_info.yml'
                result = self._get_service_info("elastic",playbook,{"action":"get_token_by_policy_name","policy_name":self.description.packetbeat_policy_name})
                agent_token = result[0]["token"]
                ansible = Ansible(deployment=self)
                inventory = ansible.build_inventory_localhost(
                    username=self.description.user_install_packetbeat,
                    extra_vars = {
                        "action": "install",
                        "elastic_url": f"https://{elastic_ip}:8220",
                        "token": agent_token,
                        "elastic_agent_version": self.description.elastic_stack_version,
                        "institution": self.description.institution,
                        "lab_name": self.description.lab_name,
                        "proxy": self.description.proxy,
                    },
                )
                ansible.run(inventory = inventory,
                    playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'agent_manage.yml',
                    quiet = True
                )
            else:
                click.echo(f"Unable to connect to Elastic. Check if machine is running.")
        except Exception as e:
            raise DeploymentDockerException(e)
    
    def _destroy_packetbeat(self, instances=None):
        ansible = Ansible(deployment=self)
        inventory = ansible.build_inventory_localhost(
            username=self.description.user_install_packetbeat,
            extra_vars = {
                "action": "delete",
                "institution": self.description.institution,
                "lab_name": self.description.lab_name,
            },
        )
        ansible.run(inventory = inventory,
            playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'agent_manage.yml',
            quiet=True
        )

    def _manage_packetbeat(self, action):
        """
        Get status of Packetbeat service.

        Returns:
          str: status of packetbeat service.
        """
        try:
            ansible = Ansible(deployment=self)
            inventory = ansible.build_inventory_localhost(
                username=self.description.user_install_packetbeat,
                extra_vars = {
                    "action": action,
                    "institution": self.description.institution,
                    "lab_name": self.description.lab_name,
                },
            )
            ansible.run(inventory = inventory,
                        playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'agent_manage.yml',
                        quiet = True)
            if action == "status":
                packetbeat_status = ansible.debug_outputs[0]["agent_status"]
                if packetbeat_status == "stopped":
                    packetbeat_status = "shutoff"
                return packetbeat_status.upper()
            else:
                return None
        except Exception as e:
            raise DeploymentDockerException(f"Unable to apply action {action} for Packetbeat. Error {e}")




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import json
from pprint import pprint

import boto3
from botocore.config import Config


class AWSClientException(Exception):
    pass


class Client:
    def __init__(self, description, connection=None, secrets_manager=None):
        """
        Initialize AWS client.

        Parameters:
            description (tectonic.description): the scenario description
            connection (boto3.client): boto3 client to use for AWS operations.
        """
        self.description = description
        if connection is not None:
            self.ec2_client = connection
        else:
            self.ec2_client = boto3.client("ec2", region_name=description.aws_region)
        if secrets_manager is not None:
            self.secretsmanager_client = secrets_manager
        else:
            self.secretsmanager_client = boto3.client(
                "secretsmanager", region_name=description.aws_region
            )

    def get_instance_property(self, instance_name, prop):
        """
        Get property of an instance given it's name.

        Parameters:
            instance_name (str): name of the instance (<institution>-<lab_name>-<instance_number>-<guest_name>-<copy>).
            prop (str): property of the instance.

        Returns:
            str: the value of the property for the instance if it is found or None otherwise.
        """
        try:
            response = self.ec2_client.describe_instances(
                Filters=[
                    {"Name": "tag:Name", "Values": [instance_name]},
                    # Filter out terminated instances
                    {
                        "Name": "instance-state-name",
                        "Values": [
                            "pending",
                            "running",
                            "shutting-down",
                            "stopping",
                            "stopped",
                        ],
                    },
                ],
                DryRun=False,
            )
            if (
                len(response["Reservations"]) > 0
                and len(response["Reservations"][0]["Instances"]) == 1
            ):
                return response["Reservations"][0]["Instances"][0].get(prop)
            return None
        except Exception as exception:
            raise AWSClientException(f"{exception}") from exception

    def start_instance(self, instance_name):
        """
        Start an instance.

        Parameters:
            instance_name (str): name of the instance (<institution>-<lab_name>-<instance_name>).
        """
        instance_id = self.get_instance_property(instance_name, "InstanceId")
        if instance_id is None:
            raise AWSClientException(f"Instance {instance_name} not found.")
        self.ec2_client.start_instances(InstanceIds=[instance_id], DryRun=False)

    def stop_instance(self, instance_name):
        """
        Stop an instance.

        Parameters:
            instance_name (str): name of the instance (<insitution>-<lab_name>-<instance_name>).
        """
        instance_id = self.get_instance_property(instance_name, "InstanceId")
        if instance_id is None:
            raise AWSClientException(f"Instance {instance_name} not found.")
        self.ec2_client.stop_instances(InstanceIds=[instance_id], DryRun=False)

    def reboot_instance(self, instance_name):
        """
        Reboot an instance.

        Parameters:
            instance_name (str): name of the instance (<institution>-<lab_name>-<instance_name>).
        """
        instance_id = self.get_instance_property(instance_name, "InstanceId")
        if instance_id is None:
            raise AWSClientException(f"Instance {instance_name} not found.")
        self.ec2_client.reboot_instances(InstanceIds=[instance_id], DryRun=False)

    def get_instance_status(self, instance_name):
        """
        Return instance status.

        Parameters:
            instance_name (str): name of the instance (<institution>-<lab_name>-<instance_name>).

        Returns:
            dict(str): dictionary with information about the state of the instance.
        """
        instance_id = self.get_instance_property(instance_name, "InstanceId")
        if instance_id is not None:
            response = self.ec2_client.describe_instance_status(
                InstanceIds=[instance_id], DryRun=False, IncludeAllInstances=True
            )
            return response["InstanceStatuses"][0]["InstanceState"]["Name"].upper()
        else:
            return "NOT FOUND"

    def get_image(self, image_name):
        """
        Get the Id of an image and its snapshot given its name.

        Parameters:
            image_name (str): name of the image (<institution>-<lab_name>-<image_name>).

        Returns:
            tuple(str): the identifier of the image an its snapshot if it was found or None otherwise.
        """
        response = self.ec2_client.describe_images(
            Filters=[{"Name": "tag:Name", "Values": [image_name]}], DryRun=False
        )
        if len(response["Images"]) == 1:
            return (
                response["Images"][0]["ImageId"],
                response["Images"][0]["BlockDeviceMappings"][0]["Ebs"]["SnapshotId"],
            )
        return None

    def delete_image(self, image_name):
        """
        Delete image and its snapshot.

        Parameters:
            image_name (str): name of the image (<institution>-<lab_name>-<image_name>).

        """
        image = self.get_image(image_name)
        if image is not None:
            self.ec2_client.deregister_image(ImageId=image[0], DryRun=False)
            self.ec2_client.delete_snapshot(SnapshotId=image[1], DryRun=False)

    def get_machine_public_ip(self, name):
        """
        Get a machine public IP.

        Parameters:
            name (str): complete name of the machine.

        Returns:
            str: machine public IP if it was found or None otherwise.
        """
        return self.get_instance_property(name, "PublicIpAddress")

    def get_machine_private_ip(self, name):
        """
        Get a machine private IP.

        Parameters:
            name (str): complete name of the machine.

        Returns:
            str : machine private IP if it was found or None otherwise.
        """
        return self.get_instance_property(name, "PrivateIpAddress")

    def get_security_group_id(self, description):
        """
        Get the Id of a security group given its description.

        Parameters:
            description (str): description of the security group.

        Returns:
            list(str): list of the identifiers of the security groups.
        """
        response = self.ec2_client.describe_security_groups(
            Filters=[{"Name": "description", "Values": [description]}],
            DryRun=False,
        )
        ids = []
        for security_group in response["SecurityGroups"]:
            ids.append(security_group["GroupId"])
        return ids

    def delete_security_groups(self, description):
        """
        Delete security groups given its description.

        Parameters:
            description (str): description of the security groups.
        """
        for sg_id in self.get_security_group_id(description):
            self.ec2_client.delete_security_group(
                GroupId=sg_id,
                DryRun=False,
            )

    def get_machines_imageid(self, max_page_results=50):
        """
        Get machines images id.

        Parameters:
            max_page_results (int): The max number of results to query in each request.

        Returns:
            list(str) : machine image ids.
        """
        try:
            first_request = True
            next_token = ""
            images_id = []
            while next_token is not None:
                if first_request:
                    response = self.ec2_client.describe_instances(
                        Filters=[
                            {
                                "Name": "instance-state-name",
                                "Values": [
                                    "pending",
                                    "running",
                                    "shutting-down",
                                    "stopping",
                                    "stopped",
                                ],
                            },
                        ],
                        DryRun=False,
                        MaxResults=max_page_results,
                    )
                else:
                    response = self.ec2_client.describe_instances(
                        Filters=[
                            {
                                "Name": "instance-state-name",
                                "Values": [
                                    "pending",
                                    "running",
                                    "shutting-down",
                                    "stopping",
                                    "stopped",
                                ],
                            },
                        ],
                        DryRun=False,
                        MaxResults=max_page_results,
                        NextToken=next_token,
                    )
                for instance in response["Reservations"]:
                    images_id.append(instance["Instances"][0]["ImageId"])
                next_token = response.get("NextToken", None)
                first_request = False
            return images_id
        except Exception as e:
            raise AWSClientException(f"{e}") from e




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.


class InstanceType:
    """This class encapsulates different machine sizes for different
    cloud providers.

    """
    
    def __init__(self, default_instance_type=None):
        self.default_instance_type = default_instance_type

    def get_guest_instance_type(self, memory, vcpus, gpu, monitor, monitor_type):
        """Default is to not have an instance type. Platform dependent
        subclasses will compute the correct instance type for the
        machine.

        """
        return self.default_instance_type




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import docker
import subprocess
from ipaddress import ip_network, ip_address


class DockerClientException(Exception):
    pass

class Client:
    
    def __init__(self, description):
        self.description = description
        try:
            self.connection = docker.DockerClient(base_url=self.description.docker_uri)
        except:
            self.connection = None
            raise DockerClientException(f"Cannot connect to docker server at {self.description.docker_uri}")

    def __del__(self):
        if self.connection is not None:
            try:
                self.connection.close()
            except:
                pass

    def get_instance_status(self, instance_name):
        try:
            container = self.connection.containers.get(instance_name)
            return container.status.upper()
        except docker.errors.NotFound:
            return "UNKNOWN"
        except Exception as exception:
            raise DockerClientException(f"{exception}")
        
    def get_machine_private_ip(self, name):
        """
        Returns the private IP address of a domain.

        If the domain has more than one IP address, the first
        address inside network_cidr_block is returned.

        """
        try:
            lab_network = ip_network(self.description.network_cidr_block)
            services_network = ip_network(self.description.services_network)
            services_list = []
            for service in self.description.get_services_to_deploy():
                services_list.append(self.description.get_service_name(service))

            container = self.connection.containers.get(name)
            for network in container.attrs["NetworkSettings"]["Networks"]:
                ip_addr = container.attrs["NetworkSettings"]["Networks"][network]["IPAddress"]
                if name in services_list:
                    if ip_address(ip_addr) in services_network:
                        return ip_addr
                else:
                    if ip_address(ip_addr) in lab_network:
                        return ip_addr
            return None #Raise exception?
        except Exception as exception:
            raise DockerClientException(f"{exception}")
        
    def delete_image(self, image_name):
        """
        Delete base image
        """
        try:
            self.connection.images.remove(image_name)
        except Exception as exception:
            raise DockerClientException(f"{exception}")
        
    def start_instance(self, instance_name):
        """
        Starts a stopped instance.
        """
        try:
            container = self.connection.containers.get(instance_name)
            container.start()
        except Exception as exception:
            raise DockerClientException(f"{exception}")

    def stop_instance(self, instance_name):
        """
        Stops a running instance.
        """
        try:
            container = self.connection.containers.get(instance_name)
            container.stop()
        except Exception as exception:
            raise DockerClientException(f"{exception}")
        
    def reboot_instance(self, instance_name):
        """
        Reboots a running instance.
        """
        try:
            container = self.connection.containers.get(instance_name)
            container.restart()
        except Exception as exception:
            raise DockerClientException(f"{exception}")   

    def connect(self, instance_name, username):
        """
        Connect to contaienr
        """
        try:
            container = self.connection.containers.get(instance_name)
            subprocess.run([f"docker exec -u {username} -it {container.id} /bin/bash"], shell=True)
            # TODO: Fix terminal using sockets
            # container = self.connection.containers.get(instance_name)
            # (_,s) = container.exec_run("/bin/bash", stdin=True, socket=True, user=username)
            # while True:
            #     original_text_to_send = input("$") + '\n'
            #     if(original_text_to_send == "exit\n"):
            #         s.close()
            #         break
            #     else:
            #         s._sock.send(original_text_to_send.encode('utf-8'))
            #         msg = s._sock.recv(1024)
            #         print(msg.decode()[8:])
        except Exception as exception:
            raise DockerClientException(f"{exception}")  
    

    
